/*
 * tut7.c -- SDL-Tutorial, 7. Kapitel
 * NO Copyright (c) by Nicolai 'Prefect' Haehnle
 *
 * Diese Software ist Public Domain.
 */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>

#include <SDL.h>


#ifndef M_PI
#define M_PI		3.14159265358979323846	/* matches value in gcc v2 math.h */
#endif


SDL_Surface *g_pSurfScreen; /* Der Bildschirminhalt */

/* Verschiedene Farbwerte */
Uint32 g_Black, g_Red, g_Blue, g_Green;

Uint32 lastframe, curframe, frametime;


/* Zeichne ein Pixel auf eine gelockte Surface */
void PutPixel(SDL_Surface *pSurf, int x, int y, Uint32 color)
{
	int bpp;
	Uint8 *p;

/* Prfe, ob das Pixel innerhalb der Surface liegt */
	if (x < 0 || x >= pSurf->w ||
	    y < 0 || y >= pSurf->h)
		return;

/* Berechne den Zielpointer */
	bpp = pSurf->format->BytesPerPixel;
	p = (Uint8 *)pSurf->pixels + y*pSurf->pitch + x*bpp;

/* Setze den Farbwert an der Stelle */
	switch(bpp) {
	case 1:
		*p = color;
		break;

	case 2:
		*(Uint16 *)p = color;
		break;

	case 4:
		*(Uint32 *)p = color;
		break;

	case 3:
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
		p[0] = (color >> 16) & 0xff;
		p[1] = (color >> 8) & 0xff;
		p[2] = color & 0xff;
#else
		p[0] = color & 0xff;
		p[1] = (color >> 8) & 0xff;
		p[2] = (color >> 16) & 0xff;
#endif
		break;
	}
}


/* Partikel-Struktur */
typedef struct particle_s {
	struct particle_s **pprev; /* Linked list */
	struct particle_s *next;
	Uint32 color; /* Farbe im Bildschirmformat */
	float x, y; /* Momentane Position auf dem Schirm */
	float vx, vy; /* Geschwindigkeit */
} particle_t;

particle_t *g_pParticles = 0; /* Linked List aller Partikel */


/* Erstelle ein neues Partikel */
particle_t *P_Create(float x, float y, Uint32 color, float vx, float vy)
{
	particle_t *p = (particle_t *)malloc(sizeof(particle_t));

/* Werte einfllen */
	p->color = color;
	p->x = x;
	p->y = y;
	p->vx = vx;
	p->vy = vy;

/* In Linked List einreihen */
	p->next = g_pParticles;
	if (p->next)
		p->next->pprev = &p->next;
	p->pprev = &g_pParticles;
	g_pParticles = p;
	return p;
}

/* Entferne ein Partikel */
void P_Remove(particle_t *p)
{
/* Aus der Linked List entfernen */
	if (p->next)
		p->next->pprev = p->pprev;
	*p->pprev = p->next;

/* Speicher freigeben */
	free(p);
}

/* Bewege alle Partikel */
void P_Think()
{
	particle_t *next = g_pParticles;

	while(next) {
		particle_t *p = next;
		next = p->next;

/* Gravitation */
		p->vy += 300 / 1000.0f * frametime;

/* Bewegung bercksichtigen */
		p->x += p->vx * frametime/1000.0f;
		p->y += p->vy * frametime/1000.0f;
		if (p->x < 0 || p->x >= 640 ||
		    p->y >= 480)
			P_Remove(p);
	}
}

/* Zeichne alle Partikel - Bildschirmsurface mu gelockt sein */
void P_Draw()
{
	particle_t *p;

	for(p = g_pParticles; p; p = p->next) {
		PutPixel(g_pSurfScreen, (int)p->x, (int)p->y, p->color);
	}
}

#define EMIT_DELAY		20 /* Millisekunden zwischen erstellten Partikeln */

Uint32 lastemit; /* Wann wurde zuletzt ein Partikel erzeugt */

/* Erstelle zufllig Partikel in der Bildschirmmitte */
void EmitParticles()
{
	while(lastemit + EMIT_DELAY < curframe) {
		float speed, angle;
		float vx, vy;
		Uint32 color;

		lastemit += EMIT_DELAY;

	/* Erstelle zufllige Werte fr Farbe und Geschwindigkeit */
		color = SDL_MapRGB(g_pSurfScreen->format, 
					(Uint8)(rand() & 0xff),
					(Uint8)(rand() & 0xff),
					(Uint8)(rand() & 0xff));
		speed = (float)(rand() % 300);
		angle = (float)(rand() * (2.0 * M_PI / RAND_MAX));
		vx = (float)(cos(angle) * speed);
		vy = (float)(sin(angle) * speed);

	/* Erstelle das Partikel */
		P_Create(320, 240, color, vx, vy);
	}
}

/* Rume auf */
void P_Exit(void)
{
	while(g_pParticles)
		P_Remove(g_pParticles);
}

/* Initialisiere das Partikelsystem */
void P_Init()
{
	atexit(P_Exit);
	lastemit = curframe;
}

#ifdef _WIN32
#undef main
#endif
int main()
{
	int running;

/* Initialisiere den Zufallszahlengenerator */
	srand(time(0));

/* SDL initialisieren */
	if (SDL_Init(SDL_INIT_VIDEO) < 0) {
		fprintf(stderr, "SDL konnte nicht initialisiert werden: %s\n",
			SDL_GetError());
		exit(1);
	}
	atexit(SDL_Quit);

/* Bildschirmmodus setzen */
	g_pSurfScreen = SDL_SetVideoMode(640, 480, 0, SDL_DOUBLEBUF);
	if (!g_pSurfScreen) {
		fprintf(stderr, "Konnte Bildschirmmodus nicht setzen: %s\n",
			SDL_GetError());
		exit(1);
	}

/* Farbwerte bestimmen */
	g_Black = SDL_MapRGB(g_pSurfScreen->format, 0, 0, 0);
	g_Red = SDL_MapRGB(g_pSurfScreen->format, 255, 0, 0);
	g_Green = SDL_MapRGB(g_pSurfScreen->format, 0, 255, 0);
	g_Blue = SDL_MapRGB(g_pSurfScreen->format, 0, 0, 255);

/* Der Timer mu initialisiert werden */
	curframe = SDL_GetTicks();

/* Programmschleife */
	P_Init();

	running = 1;
	while(running) {
		SDL_Event event;

		while(SDL_PollEvent(&event)) {
			switch(event.type) {
			case SDL_KEYDOWN:
				switch(event.key.keysym.sym) {
				/* Beim Druck von ESC beendet sich das Programm */
				case SDLK_ESCAPE:
					running = 0;
					break;

				/* Alle anderen Tasten werden ignoriert */
				default:
					break;
				}
				break;

			case SDL_QUIT:
				running = 0;
				break;
			}
		}

	/* Frametime wird berechnet */
		lastframe = curframe;
		curframe = SDL_GetTicks();
		frametime = curframe - lastframe;

	/* Bewege die Partikel */
		P_Think();
		EmitParticles();

	/* Bildschirmhintergrund lschen */
		SDL_FillRect(g_pSurfScreen, 0, g_Black);
	
		if (!SDL_LockSurface(g_pSurfScreen)) {
			P_Draw(); /* Zeichne Partikel */
			SDL_UnlockSurface(g_pSurfScreen);
		} else {
			fprintf(stderr, "SDL_LockSurface fehlgeschlagen\n");
		}

	/* Die nderung mssen wirksam gemacht werden */
		SDL_Flip(g_pSurfScreen);
	}

	return 0;
}
